import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:osam/domain/event/event.dart';
import 'package:osam/domain/middleware/middleware.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/persist/persist_repository.dart';

typedef ErrorHandler = void Function(Object error);

class Store<ST extends BaseState<ST>> with PersistRepository<ST> {
  ST state;
  final List<Middleware<ST>> middleWares;
  final ErrorHandler errorHandler;

  var logging = false;

  Stream<Event<ST>> get eventStream => _dispatcher.stream;

  // ignore: close_sinks
  final _dispatcher = StreamController<Event<ST>>.broadcast();
  final _denormalizedConditions = <Condition<ST>>[];

  Store(this.state, {this.middleWares = const [], this.errorHandler}) {
    _setupStoreForMiddlewares(middleWares);
    _denormalizeConditions();
    _dispatcher.stream.listen((event) {
      if (logging) print(event.toString());

      for (final condition in _denormalizedConditions) {
        final nextEvent = condition(event);
        if (!nextEvent) return;
      }

      try {
        event(state);
      } catch (e) {
        errorHandler?.call(e);
      }
    });
  }

  void saveState() => super.save(state);

  void restoreState() => state = super.get() ?? state;

  void dispatchEvent({@required Event<ST> event}) => _dispatcher.sink.add(event);

  void _denormalizeConditions() {
    _denormalizedConditions.clear();
    _denormalizedConditions
        .addAll(middleWares.expand((middleWare) => middleWare.conditions).toList());
  }

  void _setupStoreForMiddlewares(List<Middleware<ST>> middlewares) =>
      middlewares.forEach((middleware) => middleware.store = this);

  void updateMiddlewares(List<Middleware<ST>> middlewares) {
    _setupStoreForMiddlewares(middlewares);
    middlewares.forEach((middleware) => middleWares[middleWares.indexOf(middleware)] = middleware);
    _denormalizeConditions();
  }
}
